@@ -44,6 +44,7 @@ typedef enum fldt	{
 
 	/* attr 3 specific fields */
 	FLDT_ATTR3,
+	FLDT_ATTR3_BLKINFO,
 	FLDT_ATTR3_LEAF_HDR,
 	FLDT_ATTR3_NODE_HDR,
 
@@ -80,6 +81,8 @@ typedef enum fldt	{
 	FLDT_REXTFLG,
 	FLDT_RATTRFORKFLG,
 	FLDT_RBMBTFLG,
+	FLDT_CAGBLOCK,
+	FLDT_CCOWFLG,
 	FLDT_CNTBT,
 	FLDT_CNTBT_CRC,
 	FLDT_CNTBTKEY,
@@ -89,6 +92,10 @@ typedef enum fldt	{
 	FLDT_RMAPBTKEY,
 	FLDT_RMAPBTPTR,
 	FLDT_RMAPBTREC,
+	FLDT_REFCBT_CRC,
+	FLDT_REFCBTKEY,
+	FLDT_REFCBTPTR,
+	FLDT_REFCBTREC,
 
 	/* CRC field type */
 	FLDT_CRC,
