@@ -16,7 +16,7 @@
  * Inc.,  51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
-#include <xfs/libxfs.h>
+#include "libxfs.h"
 #include "block.h"
 #include "command.h"
 #include "type.h"
@@ -50,63 +50,120 @@ static const cmdinfo_t	type_cmd =
 	  N_("set/show current data type"), NULL };
 
 static const typ_t	__typtab[] = {
-	{ TYP_AGF, "agf", handle_struct, agf_hfld, NULL },
-	{ TYP_AGFL, "agfl", handle_struct, agfl_hfld, NULL },
-	{ TYP_AGI, "agi", handle_struct, agi_hfld, NULL },
-	{ TYP_ATTR, "attr", handle_struct, attr_hfld, NULL },
-	{ TYP_BMAPBTA, "bmapbta", handle_struct, bmapbta_hfld, NULL },
-	{ TYP_BMAPBTD, "bmapbtd", handle_struct, bmapbtd_hfld, NULL },
-	{ TYP_BNOBT, "bnobt", handle_struct, bnobt_hfld, NULL },
-	{ TYP_CNTBT, "cntbt", handle_struct, cntbt_hfld, NULL },
-	{ TYP_DATA, "data", handle_block, NULL, NULL },
-	{ TYP_DIR2, "dir2", handle_struct, dir2_hfld, NULL },
-	{ TYP_DQBLK, "dqblk", handle_struct, dqblk_hfld, NULL },
-	{ TYP_INOBT, "inobt", handle_struct, inobt_hfld, NULL },
-	{ TYP_INODATA, "inodata", NULL, NULL, NULL },
-	{ TYP_INODE, "inode", handle_struct, inode_hfld, NULL },
-	{ TYP_LOG, "log", NULL, NULL, NULL },
-	{ TYP_RTBITMAP, "rtbitmap", NULL, NULL, NULL },
-	{ TYP_RTSUMMARY, "rtsummary", NULL, NULL, NULL },
-	{ TYP_SB, "sb", handle_struct, sb_hfld, NULL },
-	{ TYP_SYMLINK, "symlink", handle_string, NULL, NULL },
-	{ TYP_TEXT, "text", handle_text, NULL, NULL },
-	{ TYP_FINOBT, "finobt", handle_struct, inobt_hfld, NULL },
+	{ TYP_AGF, "agf", handle_struct, agf_hfld, NULL, TYP_F_NO_CRC_OFF },
+	{ TYP_AGFL, "agfl", handle_struct, agfl_hfld, NULL, TYP_F_NO_CRC_OFF },
+	{ TYP_AGI, "agi", handle_struct, agi_hfld, NULL, TYP_F_NO_CRC_OFF },
+	{ TYP_ATTR, "attr", handle_struct, attr_hfld, NULL, TYP_F_NO_CRC_OFF },
+	{ TYP_BMAPBTA, "bmapbta", handle_struct, bmapbta_hfld, NULL,
+		TYP_F_NO_CRC_OFF },
+	{ TYP_BMAPBTD, "bmapbtd", handle_struct, bmapbtd_hfld, NULL,
+		TYP_F_NO_CRC_OFF },
+	{ TYP_BNOBT, "bnobt", handle_struct, bnobt_hfld, NULL, TYP_F_NO_CRC_OFF },
+	{ TYP_CNTBT, "cntbt", handle_struct, cntbt_hfld, NULL, TYP_F_NO_CRC_OFF },
+	{ TYP_RMAPBT, NULL },
+	{ TYP_REFCBT, NULL },
+	{ TYP_DATA, "data", handle_block, NULL, NULL, TYP_F_NO_CRC_OFF },
+	{ TYP_DIR2, "dir2", handle_struct, dir2_hfld, NULL, TYP_F_NO_CRC_OFF },
+	{ TYP_DQBLK, "dqblk", handle_struct, dqblk_hfld, NULL, TYP_F_NO_CRC_OFF },
+	{ TYP_INOBT, "inobt", handle_struct, inobt_hfld, NULL, TYP_F_NO_CRC_OFF },
+	{ TYP_INODATA, "inodata", NULL, NULL, NULL, TYP_F_NO_CRC_OFF },
+	{ TYP_INODE, "inode", handle_struct, inode_hfld, NULL, TYP_F_NO_CRC_OFF },
+	{ TYP_LOG, "log", NULL, NULL, NULL, TYP_F_NO_CRC_OFF },
+	{ TYP_RTBITMAP, "rtbitmap", NULL, NULL, NULL, TYP_F_NO_CRC_OFF },
+	{ TYP_RTSUMMARY, "rtsummary", NULL, NULL, NULL, TYP_F_NO_CRC_OFF },
+	{ TYP_SB, "sb", handle_struct, sb_hfld, NULL, TYP_F_NO_CRC_OFF },
+	{ TYP_SYMLINK, "symlink", handle_string, NULL, NULL, TYP_F_NO_CRC_OFF },
+	{ TYP_TEXT, "text", handle_text, NULL, NULL, TYP_F_NO_CRC_OFF },
+	{ TYP_FINOBT, "finobt", handle_struct, inobt_hfld, NULL,
+		TYP_F_NO_CRC_OFF },
 	{ TYP_NONE, NULL }
 };
 
 static const typ_t	__typtab_crc[] = {
-	{ TYP_AGF, "agf", handle_struct, agf_hfld, &xfs_agf_buf_ops },
-	{ TYP_AGFL, "agfl", handle_struct, agfl_crc_hfld, &xfs_agfl_buf_ops },
-	{ TYP_AGI, "agi", handle_struct, agi_hfld, &xfs_agfl_buf_ops },
+	{ TYP_AGF, "agf", handle_struct, agf_hfld, &xfs_agf_buf_ops,
+		XFS_AGF_CRC_OFF },
+	{ TYP_AGFL, "agfl", handle_struct, agfl_crc_hfld, &xfs_agfl_buf_ops,
+		XFS_AGFL_CRC_OFF },
+	{ TYP_AGI, "agi", handle_struct, agi_hfld, &xfs_agi_buf_ops,
+		XFS_AGI_CRC_OFF },
 	{ TYP_ATTR, "attr3", handle_struct, attr3_hfld,
-		&xfs_attr3_db_buf_ops },
+		&xfs_attr3_db_buf_ops, TYP_F_NO_CRC_OFF },
 	{ TYP_BMAPBTA, "bmapbta", handle_struct, bmapbta_crc_hfld,
-		&xfs_bmbt_buf_ops },
+		&xfs_bmbt_buf_ops, XFS_BTREE_LBLOCK_CRC_OFF },
 	{ TYP_BMAPBTD, "bmapbtd", handle_struct, bmapbtd_crc_hfld,
-		&xfs_bmbt_buf_ops },
+		&xfs_bmbt_buf_ops, XFS_BTREE_LBLOCK_CRC_OFF },
 	{ TYP_BNOBT, "bnobt", handle_struct, bnobt_crc_hfld,
-		&xfs_allocbt_buf_ops },
+		&xfs_allocbt_buf_ops, XFS_BTREE_SBLOCK_CRC_OFF },
 	{ TYP_CNTBT, "cntbt", handle_struct, cntbt_crc_hfld,
-		&xfs_allocbt_buf_ops },
-	{ TYP_DATA, "data", handle_block, NULL, NULL },
+		&xfs_allocbt_buf_ops, XFS_BTREE_SBLOCK_CRC_OFF },
+	{ TYP_RMAPBT, "rmapbt", handle_struct, rmapbt_crc_hfld,
+		&xfs_rmapbt_buf_ops, XFS_BTREE_SBLOCK_CRC_OFF },
+	{ TYP_REFCBT, "refcntbt", handle_struct, refcbt_crc_hfld,
+		&xfs_refcountbt_buf_ops, XFS_BTREE_SBLOCK_CRC_OFF },
+	{ TYP_DATA, "data", handle_block, NULL, NULL, TYP_F_NO_CRC_OFF },
 	{ TYP_DIR2, "dir3", handle_struct, dir3_hfld,
-		&xfs_dir3_db_buf_ops },
+		&xfs_dir3_db_buf_ops, TYP_F_NO_CRC_OFF },
 	{ TYP_DQBLK, "dqblk", handle_struct, dqblk_hfld,
-		&xfs_dquot_buf_ops },
+		&xfs_dquot_buf_ops, TYP_F_NO_CRC_OFF },
 	{ TYP_INOBT, "inobt", handle_struct, inobt_crc_hfld,
-		&xfs_inobt_buf_ops },
-	{ TYP_INODATA, "inodata", NULL, NULL, NULL },
+		&xfs_inobt_buf_ops, XFS_BTREE_SBLOCK_CRC_OFF },
+	{ TYP_INODATA, "inodata", NULL, NULL, NULL, TYP_F_NO_CRC_OFF },
 	{ TYP_INODE, "inode", handle_struct, inode_crc_hfld,
-		&xfs_inode_buf_ops },
-	{ TYP_LOG, "log", NULL, NULL, NULL },
-	{ TYP_RTBITMAP, "rtbitmap", NULL, NULL, NULL },
-	{ TYP_RTSUMMARY, "rtsummary", NULL, NULL, NULL },
-	{ TYP_SB, "sb", handle_struct, sb_hfld, &xfs_sb_buf_ops },
+		&xfs_inode_buf_ops, TYP_F_NO_CRC_OFF },
+	{ TYP_LOG, "log", NULL, NULL, NULL, TYP_F_NO_CRC_OFF },
+	{ TYP_RTBITMAP, "rtbitmap", NULL, NULL, NULL, TYP_F_NO_CRC_OFF },
+	{ TYP_RTSUMMARY, "rtsummary", NULL, NULL, NULL, TYP_F_NO_CRC_OFF },
+	{ TYP_SB, "sb", handle_struct, sb_hfld, &xfs_sb_buf_ops,
+		XFS_SB_CRC_OFF },
 	{ TYP_SYMLINK, "symlink", handle_struct, symlink_crc_hfld,
-		&xfs_symlink_buf_ops },
-	{ TYP_TEXT, "text", handle_text, NULL, NULL },
+		&xfs_symlink_buf_ops, XFS_SYMLINK_CRC_OFF },
+	{ TYP_TEXT, "text", handle_text, NULL, NULL, TYP_F_NO_CRC_OFF },
 	{ TYP_FINOBT, "finobt", handle_struct, inobt_crc_hfld,
-		&xfs_inobt_buf_ops },
+		&xfs_inobt_buf_ops, XFS_BTREE_SBLOCK_CRC_OFF },
+	{ TYP_NONE, NULL }
+};
+
+static const typ_t	__typtab_spcrc[] = {
+	{ TYP_AGF, "agf", handle_struct, agf_hfld, &xfs_agf_buf_ops,
+		XFS_AGF_CRC_OFF },
+	{ TYP_AGFL, "agfl", handle_struct, agfl_crc_hfld, &xfs_agfl_buf_ops ,
+		XFS_AGFL_CRC_OFF },
+	{ TYP_AGI, "agi", handle_struct, agi_hfld, &xfs_agi_buf_ops ,
+		XFS_AGI_CRC_OFF },
+	{ TYP_ATTR, "attr3", handle_struct, attr3_hfld,
+		&xfs_attr3_db_buf_ops, TYP_F_NO_CRC_OFF },
+	{ TYP_BMAPBTA, "bmapbta", handle_struct, bmapbta_crc_hfld,
+		&xfs_bmbt_buf_ops, XFS_BTREE_LBLOCK_CRC_OFF },
+	{ TYP_BMAPBTD, "bmapbtd", handle_struct, bmapbtd_crc_hfld,
+		&xfs_bmbt_buf_ops, XFS_BTREE_LBLOCK_CRC_OFF },
+	{ TYP_BNOBT, "bnobt", handle_struct, bnobt_crc_hfld,
+		&xfs_allocbt_buf_ops, XFS_BTREE_SBLOCK_CRC_OFF },
+	{ TYP_CNTBT, "cntbt", handle_struct, cntbt_crc_hfld,
+		&xfs_allocbt_buf_ops, XFS_BTREE_SBLOCK_CRC_OFF },
+	{ TYP_RMAPBT, "rmapbt", handle_struct, rmapbt_crc_hfld,
+		&xfs_rmapbt_buf_ops, XFS_BTREE_SBLOCK_CRC_OFF },
+	{ TYP_REFCBT, "refcntbt", handle_struct, refcbt_crc_hfld,
+		&xfs_refcountbt_buf_ops, XFS_BTREE_SBLOCK_CRC_OFF },
+	{ TYP_DATA, "data", handle_block, NULL, NULL, TYP_F_NO_CRC_OFF },
+	{ TYP_DIR2, "dir3", handle_struct, dir3_hfld,
+		&xfs_dir3_db_buf_ops, TYP_F_NO_CRC_OFF },
+	{ TYP_DQBLK, "dqblk", handle_struct, dqblk_hfld,
+		&xfs_dquot_buf_ops, TYP_F_NO_CRC_OFF },
+	{ TYP_INOBT, "inobt", handle_struct, inobt_spcrc_hfld,
+		&xfs_inobt_buf_ops, XFS_BTREE_SBLOCK_CRC_OFF },
+	{ TYP_INODATA, "inodata", NULL, NULL, NULL, TYP_F_NO_CRC_OFF },
+	{ TYP_INODE, "inode", handle_struct, inode_crc_hfld,
+		&xfs_inode_buf_ops, TYP_F_NO_CRC_OFF },
+	{ TYP_LOG, "log", NULL, NULL, NULL, TYP_F_NO_CRC_OFF },
+	{ TYP_RTBITMAP, "rtbitmap", NULL, NULL, NULL, TYP_F_NO_CRC_OFF },
+	{ TYP_RTSUMMARY, "rtsummary", NULL, NULL, NULL, TYP_F_NO_CRC_OFF },
+	{ TYP_SB, "sb", handle_struct, sb_hfld, &xfs_sb_buf_ops,
+		XFS_SB_CRC_OFF },
+	{ TYP_SYMLINK, "symlink", handle_struct, symlink_crc_hfld,
+		&xfs_symlink_buf_ops, XFS_SYMLINK_CRC_OFF },
+	{ TYP_TEXT, "text", handle_text, NULL, NULL, TYP_F_NO_CRC_OFF },
+	{ TYP_FINOBT, "finobt", handle_struct, inobt_crc_hfld,
+		&xfs_inobt_buf_ops, XFS_BTREE_SBLOCK_CRC_OFF },
 	{ TYP_NONE, NULL }
 };
 
@@ -118,6 +175,12 @@ type_set_tab_crc(void)
 	typtab = __typtab_crc;
 }
 
+void
+type_set_tab_spcrc(void)
+{
+	typtab = __typtab_spcrc;
+}
+
 static const typ_t *
 findtyp(
 	char		*name)
